@@ -12,7 +12,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import os, sys
+import os
 import pickle
 import xml.dom.minidom
 import xml.etree.ElementTree as ET
@@ -227,7 +227,7 @@ class Vs2010Backend(backends.Backend):
 
     def generate_solution(self, sln_filename, projlist):
         default_projlist = self.get_build_by_default_targets()
-        with open(sln_filename, 'w') as ofile:
+        with open(sln_filename, 'w', encoding='utf-8') as ofile:
             ofile.write('Microsoft Visual Studio Solution File, Format '
                         'Version 11.00\n')
             ofile.write('# Visual Studio ' + self.vs_version + '\n')
@@ -304,6 +304,7 @@ class Vs2010Backend(backends.Backend):
         projlist = []
         for name, target in self.build.targets.items():
             outdir = os.path.join(self.environment.get_build_dir(), self.get_target_dir(target))
+            os.makedirs(outdir, exist_ok=True)
             fname = name + '.vcxproj'
             relname = os.path.join(target.subdir, fname)
             projfile = os.path.join(outdir, fname)
@@ -574,7 +575,7 @@ class Vs2010Backend(backends.Backend):
         tree.write(ofname, encoding='utf-8', xml_declaration=True)
         # ElementTree can not do prettyprinting so do it manually
         doc = xml.dom.minidom.parse(ofname)
-        with open(ofname, 'w') as of:
+        with open(ofname, 'w', encoding='utf-8') as of:
             of.write(doc.toprettyxml())
 
     def gen_vcxproj(self, target, ofname, guid):
@@ -730,7 +731,7 @@ class Vs2010Backend(backends.Backend):
         # generate_single_compile() and generate_basic_compiler_args()
         for l, comp in target.compilers.items():
             if l in file_args:
-                file_args[l] += compilers.get_base_compile_args(self.environment.coredata.base_options, comp)
+                file_args[l] += compilers.get_base_compile_args(self.get_base_options_for_target(target), comp)
                 file_args[l] += comp.get_option_compile_args(self.environment.coredata.compiler_options)
         # Add compile args added using add_project_arguments()
         for l, args in self.build.projects_args.get(target.subproject, {}).items():
@@ -768,7 +769,8 @@ class Vs2010Backend(backends.Backend):
             # These are per-target, but we still add them as per-file because we
             # need them to be looked in first.
             for d in reversed(target.get_include_dirs()):
-                for i in d.get_incdirs():
+                # reversed is used to keep order of includes
+                for i in reversed(d.get_incdirs()):
                     curdir = os.path.join(d.get_curdir(), i)
                     args.append('-I' + self.relpath(curdir, target.subdir)) # build dir
                     args.append('-I' + os.path.join(proj_to_src_root, curdir)) # src dir
@@ -822,7 +824,10 @@ class Vs2010Backend(backends.Backend):
         for d in reversed(target.get_external_deps()):
             # Cflags required by external deps might have UNIX-specific flags,
             # so filter them out if needed
-            d_compile_args = compiler.unix_args_to_native(d.get_compile_args())
+            if isinstance(d, dependencies.OpenMPDependency):
+                d_compile_args = compiler.openmp_flags()
+            else:
+                d_compile_args = compiler.unix_args_to_native(d.get_compile_args())
             for arg in d_compile_args:
                 if arg.startswith(('-D', '/D')):
                     define = arg[2:]
@@ -913,11 +918,17 @@ class Vs2010Backend(backends.Backend):
             for dep in target.get_external_deps():
                 # Extend without reordering or de-dup to preserve `-L -l` sets
                 # https://github.com/mesonbuild/meson/issues/1718
-                extra_link_args.extend_direct(dep.get_link_args())
+                if isinstance(dep, dependencies.OpenMPDependency):
+                    extra_link_args.extend_direct(compiler.openmp_flags())
+                else:
+                    extra_link_args.extend_direct(dep.get_link_args())
             for d in target.get_dependencies():
                 if isinstance(d, build.StaticLibrary):
                     for dep in d.get_external_deps():
-                        extra_link_args.extend_direct(dep.get_link_args())
+                        if isinstance(dep, dependencies.OpenMPDependency):
+                            extra_link_args.extend_direct(compiler.openmp_flags())
+                        else:
+                            extra_link_args.extend_direct(dep.get_link_args())
         # Add link args for c_* or cpp_* build options. Currently this only
         # adds c_winlibs and cpp_winlibs when building for Windows. This needs
         # to be after all internal and external libraries so that unresolved
@@ -944,7 +955,8 @@ class Vs2010Backend(backends.Backend):
                 self.add_project_reference(root, tvcxproj, tid)
             else:
                 # Other libraries go into AdditionalDependencies
-                additional_links.append(linkname)
+                if linkname not in additional_links:
+                    additional_links.append(linkname)
         for lib in self.get_custom_target_provided_libraries(target):
             additional_links.append(self.relpath(lib, self.get_target_dir(target)))
         additional_objects = []
@@ -1014,7 +1026,7 @@ class Vs2010Backend(backends.Backend):
                 self.add_additional_options(lang, inc_cl, file_args)
                 self.add_preprocessor_defines(lang, inc_cl, file_defines)
                 self.add_include_dirs(lang, inc_cl, file_inc_dirs)
-                ET.SubElement(inc_cl, 'ObjectFileName').text = "$(IntDir)" + self.object_filename_from_source(target, s, False)
+                ET.SubElement(inc_cl, 'ObjectFileName').text = "$(IntDir)" + self.object_filename_from_source(target, s)
             for s in gen_src:
                 inc_cl = ET.SubElement(inc_src, 'CLCompile', Include=s)
                 lang = Vs2010Backend.lang_from_source_file(s)
@@ -1116,7 +1128,7 @@ if %%errorlevel%% neq 0 goto :VCEnd'''
         igroup = ET.SubElement(root, 'ItemGroup')
         rulefile = os.path.join(self.environment.get_scratch_dir(), 'regen.rule')
         if not os.path.exists(rulefile):
-            with open(rulefile, 'w') as f:
+            with open(rulefile, 'w', encoding='utf-8') as f:
                 f.write("# Meson regen file.")
         custombuild = ET.SubElement(igroup, 'CustomBuild', Include=rulefile)
         message = ET.SubElement(custombuild, 'Message')
