@@ -44,6 +44,8 @@ from .compilers import (
 
 
 class CCompiler(Compiler):
+    library_dirs_cache = {}
+
     def __init__(self, exelist, version, is_cross, exe_wrapper=None, **kwargs):
         # If a child ObjC or CPP class has already set it, don't set it ourselves
         if not hasattr(self, 'language'):
@@ -57,6 +59,9 @@ class CCompiler(Compiler):
         else:
             self.exe_wrapper = exe_wrapper
 
+        # Set to None until we actually need to check this
+        self.has_fatal_warnings_link_arg = None
+
     def needs_static_linker(self):
         return True # When compiling static libraries, so yes.
 
@@ -86,7 +91,7 @@ class CCompiler(Compiler):
         # Almost every compiler uses this for disabling warnings
         return ['-w']
 
-    def get_soname_args(self, prefix, shlib_name, suffix, path, soversion, version, is_shared_module):
+    def get_soname_args(self, prefix, shlib_name, suffix, path, soversion, is_shared_module):
         return []
 
     def split_shlib_to_parts(self, fname):
@@ -157,7 +162,7 @@ class CCompiler(Compiler):
     def get_std_shared_lib_link_args(self):
         return ['-shared']
 
-    def get_library_dirs(self):
+    def get_library_dirs_real(self):
         env = os.environ.copy()
         env['LC_ALL'] = 'C'
         stdo = Popen_safe(self.exelist + ['--print-search-dirs'], env=env)[1]
@@ -167,6 +172,12 @@ class CCompiler(Compiler):
                 return libstr.split(':')
         return []
 
+    def get_library_dirs(self):
+        key = tuple(self.exelist)
+        if key not in self.library_dirs_cache:
+            self.library_dirs_cache[key] = self.get_library_dirs_real()
+        return self.library_dirs_cache[key][:]
+
     def get_pic_args(self):
         return ['-fPIC']
 
@@ -863,6 +874,17 @@ class CCompiler(Compiler):
         return self.has_arguments(args, env, code, mode='compile')
 
     def has_multi_link_arguments(self, args, env):
+        # First time we check for link flags we need to first check if we have
+        # --fatal-warnings, otherwise some linker checks could give some
+        # false positive.
+        fatal_warnings_args = ['-Wl,--fatal-warnings']
+        if self.has_fatal_warnings_link_arg is None:
+            self.has_fatal_warnings_link_arg = False
+            self.has_fatal_warnings_link_arg = self.has_multi_link_arguments(fatal_warnings_args, env)
+
+        if self.has_fatal_warnings_link_arg:
+            args = fatal_warnings_args + args
+
         args = self.linker_to_compiler_args(args)
         code = 'int main(int argc, char **argv) { return 0; }'
         return self.has_arguments(args, env, code, mode='link')
