@@ -21,6 +21,21 @@ import platform, subprocess, operator, os, shutil, re
 import collections
 from mesonbuild import mlog
 
+have_fcntl = False
+have_msvcrt = False
+
+try:
+    import fcntl
+    have_fcntl = True
+except Exception:
+    pass
+
+try:
+    import msvcrt
+    have_msvcrt = True
+except Exception:
+    pass
+
 from glob import glob
 
 def detect_meson_py_location():
@@ -293,6 +308,9 @@ def is_android():
 def is_haiku():
     return platform.system().lower() == 'haiku'
 
+def is_openbsd():
+    return platform.system().lower() == 'openbsd'
+
 def is_windows():
     platname = platform.system().lower()
     return platname == 'windows' or 'mingw' in platname
@@ -519,9 +537,20 @@ def get_library_dirs():
     unixdirs += glob('/lib/' + plat + '*')
     return unixdirs
 
+def has_path_sep(name, sep='/\\'):
+    'Checks if any of the specified @sep path separators are in @name'
+    for each in sep:
+        if each in name:
+            return True
+    return False
 
-def do_replacement(regex, line, confdata):
+def do_replacement(regex, line, format, confdata):
     missing_variables = set()
+    start_tag = '@'
+    backslash_tag = '\\@'
+    if format == 'cmake':
+        start_tag = '${'
+        backslash_tag = '\\${'
 
     def variable_replace(match):
         # Pairs of escape characters before '@' or '\@'
@@ -529,8 +558,8 @@ def do_replacement(regex, line, confdata):
             num_escapes = match.end(0) - match.start(0)
             return '\\' * (num_escapes // 2)
         # Single escape character and '@'
-        elif match.group(0) == '\\@':
-            return '@'
+        elif match.group(0) == backslash_tag:
+            return start_tag
         # Template variable to be replaced
         else:
             varname = match.group(1)
@@ -570,7 +599,7 @@ def do_mesondefine(line, confdata):
         raise MesonException('#mesondefine argument "%s" is of unknown type.' % varname)
 
 
-def do_conf_file(src, dst, confdata):
+def do_conf_file(src, dst, confdata, format):
     try:
         with open(src, encoding='utf-8') as f:
             data = f.readlines()
@@ -578,14 +607,24 @@ def do_conf_file(src, dst, confdata):
         raise MesonException('Could not read input file %s: %s' % (src, str(e)))
     # Only allow (a-z, A-Z, 0-9, _, -) as valid characters for a define
     # Also allow escaping '@' with '\@'
-    regex = re.compile(r'(?:\\\\)+(?=\\?@)|\\@|@([-a-zA-Z0-9_]+)@')
+    if format in ['meson', 'cmake@']:
+        regex = re.compile(r'(?:\\\\)+(?=\\?@)|\\@|@([-a-zA-Z0-9_]+)@')
+    elif format == 'cmake':
+        regex = re.compile(r'(?:\\\\)+(?=\\?\$)|\\\${|\${([-a-zA-Z0-9_]+)}')
+    else:
+        raise MesonException('Format "{}" not handled'.format(format))
+
+    search_token = '#mesondefine'
+    if format != 'meson':
+        search_token = '#cmakedefine'
+
     result = []
     missing_variables = set()
     for line in data:
-        if line.startswith('#mesondefine'):
+        if line.startswith(search_token):
             line = do_mesondefine(line, confdata)
         else:
-            line, missing = do_replacement(regex, line, confdata)
+            line, missing = do_replacement(regex, line, format, confdata)
             missing_variables.update(missing)
         result.append(line)
     dst_tmp = dst + '~'
@@ -713,7 +752,9 @@ def expand_arguments(args):
     return expended_args
 
 def Popen_safe(args, write=None, stderr=subprocess.PIPE, **kwargs):
-    if sys.version_info < (3, 6) or not sys.stdout.encoding:
+    import locale
+    encoding = locale.getpreferredencoding()
+    if sys.version_info < (3, 6) or not sys.stdout.encoding or encoding.upper() != 'UTF-8':
         return Popen_safe_legacy(args, write=write, stderr=stderr, **kwargs)
     p = subprocess.Popen(args, universal_newlines=True,
                          close_fds=False,
@@ -972,3 +1013,26 @@ class OrderedSet(collections.MutableSet):
 
     def difference(self, set_):
         return type(self)(e for e in self if e not in set_)
+
+class BuildDirLock:
+
+    def __init__(self, builddir):
+        self.lockfilename = os.path.join(builddir, 'meson-private/meson.lock')
+
+    def __enter__(self):
+        self.lockfile = open(self.lockfilename, 'w')
+        try:
+            if have_fcntl:
+                fcntl.flock(self.lockfile, fcntl.LOCK_EX | fcntl.LOCK_NB)
+            elif have_msvcrt:
+                msvcrt.locking(self.lockfile.fileno(), msvcrt.LK_NBLCK, 1)
+        except (BlockingIOError, PermissionError):
+            self.lockfile.close()
+            raise MesonException('Some other Meson process is already using this build directory. Exiting.')
+
+    def __exit__(self, *args):
+        if have_fcntl:
+            fcntl.flock(self.lockfile, fcntl.LOCK_UN)
+        elif have_msvcrt:
+            msvcrt.locking(self.lockfile.fileno(), msvcrt.LK_UNLCK, 1)
+        self.lockfile.close()
