@@ -15,10 +15,12 @@
 import sys, stat, traceback, argparse
 import datetime
 import os.path
+import platform
+import cProfile as profile
+
 from . import environment, interpreter, mesonlib
 from . import build
 from . import mconf, mintro, mtest, rewriter, minit
-import platform
 from . import mlog, coredata
 from .mesonlib import MesonException
 from .wrap import WrapMode, wraptool
@@ -30,14 +32,14 @@ def create_parser():
     coredata.register_builtin_arguments(p)
     p.add_argument('--cross-file', default=None,
                    help='File describing cross compilation environment.')
-    p.add_argument('-D', action='append', dest='projectoptions', default=[], metavar="option",
-                   help='Set the value of an option, can be used several times to set multiple options.')
     p.add_argument('-v', '--version', action='version',
                    version=coredata.version)
     # See the mesonlib.WrapMode enum for documentation
     p.add_argument('--wrap-mode', default=WrapMode.default,
                    type=wrapmodetype, choices=WrapMode,
                    help='Special wrap mode to use')
+    p.add_argument('--profile-self', action='store_true', dest='profile',
+                   help=argparse.SUPPRESS)
     p.add_argument('directories', nargs='*')
     return p
 
@@ -59,11 +61,12 @@ def filter_builtin_options(args, original_args):
     if meson_opts:
         for arg in meson_opts:
             value = arguments[arg]
-            if any([a.startswith('--{}'.format(arg)) for a in original_args]):
+            cmdline_name = coredata.get_builtin_option_cmdline_name(arg)
+            if any([a.startswith(cmdline_name) for a in original_args]):
                 raise MesonException(
-                    'Argument "{0}" passed as both --{0} and -D{0}, but only '
-                    'one is allowed'.format(arg))
-            setattr(args, coredata.get_builtin_option_destination(arg), value)
+                    'Argument "{0}" passed as both {1} and -D{0}, but only '
+                    'one is allowed'.format(arg, cmdline_name))
+            setattr(args, arg, value)
 
             # Remove the builtin option from the project args values
             args.projectoptions.remove('{}={}'.format(arg, value))
@@ -176,7 +179,11 @@ class MesonApp:
             mlog.log('Target machine cpu:', mlog.bold(intr.builtin['target_machine'].cpu_method([], {})))
         mlog.log('Build machine cpu family:', mlog.bold(intr.builtin['build_machine'].cpu_family_method([], {})))
         mlog.log('Build machine cpu:', mlog.bold(intr.builtin['build_machine'].cpu_method([], {})))
-        intr.run()
+        if self.options.profile:
+            fname = os.path.join(self.build_dir, 'meson-private', 'profile-interpreter.log')
+            profile.runctx('intr.run()', globals(), locals(), filename=fname)
+        else:
+            intr.run()
         try:
             dumpfile = os.path.join(env.get_scratch_dir(), 'build.dat')
             # We would like to write coredata as late as possible since we use the existence of
@@ -186,7 +193,12 @@ class MesonApp:
             # sync with the time that gets applied to any files. Thus, we dump this file as late as
             # possible, but before build files, and if any error occurs, delete it.
             cdf = env.dump_coredata()
-            g.generate(intr)
+            if self.options.profile:
+                fname = 'profile-{}-backend.log'.format(self.options.backend)
+                fname = os.path.join(self.build_dir, 'meson-private', fname)
+                profile.runctx('g.generate(intr)', globals(), locals(), filename=fname)
+            else:
+                g.generate(intr)
             build.save(b, dumpfile)
             # Post-conf scripts must be run after writing coredata or else introspection fails.
             g.run_postconf_scripts()
