@@ -13,6 +13,7 @@
 # limitations under the License.
 
 import sys, os, platform, io
+from contextlib import contextmanager
 
 """This is (mostly) a standalone module used to write logging
 information about Meson runs. Some output goes to screen,
@@ -25,6 +26,7 @@ else:
 log_dir = None
 log_file = None
 log_fname = 'meson-log.txt'
+log_depth = 0
 
 def initialize(logdir):
     global log_dir, log_file
@@ -77,15 +79,21 @@ def process_markup(args, keep):
     return arr
 
 def force_print(*args, **kwargs):
+    iostr = io.StringIO()
+    kwargs['file'] = iostr
+    print(*args, **kwargs)
+
+    raw = iostr.getvalue()
+    if log_depth > 0:
+        prepend = '|' * log_depth
+        raw = prepend + raw.replace('\n', '\n' + prepend, raw.count('\n') - 1)
+
     # _Something_ is going to get printed.
     try:
-        print(*args, **kwargs)
+        print(raw, end='')
     except UnicodeEncodeError:
-        iostr = io.StringIO()
-        kwargs['file'] = iostr
-        print(*args, **kwargs)
-        cleaned = iostr.getvalue().encode('ascii', 'replace').decode('ascii')
-        print(cleaned)
+        cleaned = raw.encode('ascii', 'replace').decode('ascii')
+        print(cleaned, end='')
 
 def debug(*args, **kwargs):
     arr = process_markup(args, False)
@@ -102,8 +110,37 @@ def log(*args, **kwargs):
         arr = process_markup(args, True)
     force_print(*arr, **kwargs)
 
+def _log_error(severity, *args, **kwargs):
+    from . import environment
+    if severity == 'warning':
+        args = (yellow('WARNING:'),) + args
+    elif severity == 'error':
+        args = (red('ERROR:'),) + args
+    else:
+        assert False, 'Invalid severity ' + severity
+
+    if 'location' in kwargs:
+        location = kwargs['location']
+        del kwargs['location']
+        location_str = '{}:{}:'.format(os.path.join(location.subdir,
+                                                    environment.build_filename),
+                                       location.lineno)
+        args = (location_str,) + args
+
+    log(*args, **kwargs)
+
+def error(*args, **kwargs):
+    return _log_error('error', *args, **kwargs)
+
 def warning(*args, **kwargs):
-    log(yellow('WARNING:'), *args, **kwargs)
+    return _log_error('warning', *args, **kwargs)
+
+def exception(e):
+    log()
+    if hasattr(e, 'file') and hasattr(e, 'lineno') and hasattr(e, 'colno'):
+        log('%s:%d:%d:' % (e.file, e.lineno, e.colno), red('ERROR: '), e)
+    else:
+        log(red('ERROR:'), e)
 
 # Format a list for logging purposes as a string. It separates
 # all but the last item with commas, and the last with 'and'.
@@ -117,3 +154,12 @@ def format_list(list):
         return list[0]
     else:
         return ''
+
+@contextmanager
+def nested():
+    global log_depth
+    log_depth += 1
+    try:
+        yield
+    finally:
+        log_depth -= 1
