@@ -97,6 +97,10 @@ def restore_selinux_contexts():
         # is ignored quietly.
         return
 
+    if not shutil.which('restorecon'):
+        # If we don't have restorecon, failure is ignored quietly.
+        return
+
     with subprocess.Popen(['restorecon', '-F', '-f-', '-0'],
                           stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as proc:
         out, err = proc.communicate(input=b'\0'.join(os.fsencode(f) for f in selinux_updates) + b'\0')
@@ -128,20 +132,42 @@ def do_copyfile(from_file, to_file):
     selinux_updates.append(to_file)
     append_to_log(to_file)
 
-def do_copydir(data, src_prefix, src_dir, dst_dir, exclude):
+def do_copydir(data, src_dir, dst_dir, exclude):
     '''
-    Copies the directory @src_prefix (full path) into @dst_dir
+    Copies the contents of directory @src_dir into @dst_dir.
 
-    @src_dir is simply the parent directory of @src_prefix
+    For directory
+        /foo/
+          bar/
+            excluded
+            foobar
+          file
+    do_copydir(..., '/foo', '/dst/dir', {'bar/excluded'}) creates
+        /dst/
+          dir/
+            bar/
+              foobar
+            file
+
+    Args:
+        src_dir: str, absolute path to the source directory
+        dst_dir: str, absolute path to the destination directory
+        exclude: (set(str), set(str)), tuple of (exclude_files, exclude_dirs),
+                 each element of the set is a path relative to src_dir.
     '''
+    if not os.path.isabs(src_dir):
+        raise ValueError('src_dir must be absolute, got %s' % src_dir)
+    if not os.path.isabs(dst_dir):
+        raise ValueError('dst_dir must be absolute, got %s' % dst_dir)
     if exclude is not None:
         exclude_files, exclude_dirs = exclude
     else:
         exclude_files = exclude_dirs = set()
-    for root, dirs, files in os.walk(src_prefix):
+    for root, dirs, files in os.walk(src_dir):
+        assert os.path.isabs(root)
         for d in dirs[:]:
-            abs_src = os.path.join(src_dir, root, d)
-            filepart = abs_src[len(src_dir) + 1:]
+            abs_src = os.path.join(root, d)
+            filepart = os.path.relpath(abs_src, start=src_dir)
             abs_dst = os.path.join(dst_dir, filepart)
             # Remove these so they aren't visited by os.walk at all.
             if filepart in exclude_dirs:
@@ -155,8 +181,8 @@ def do_copydir(data, src_prefix, src_dir, dst_dir, exclude):
             data.dirmaker.makedirs(abs_dst)
             shutil.copystat(abs_src, abs_dst)
         for f in files:
-            abs_src = os.path.join(src_dir, root, f)
-            filepart = abs_src[len(src_dir) + 1:]
+            abs_src = os.path.join(root, f)
+            filepart = os.path.relpath(abs_src, start=src_dir)
             if filepart in exclude_files:
                 continue
             abs_dst = os.path.join(dst_dir, filepart)
@@ -164,10 +190,10 @@ def do_copydir(data, src_prefix, src_dir, dst_dir, exclude):
                 print('Tried to copy file %s but a directory of that name already exists.' % abs_dst)
             if os.path.exists(abs_dst):
                 os.unlink(abs_dst)
-            parent_dir = os.path.split(abs_dst)[0]
+            parent_dir = os.path.dirname(abs_dst)
             if not os.path.isdir(parent_dir):
                 os.mkdir(parent_dir)
-                shutil.copystat(os.path.split(abs_src)[0], parent_dir)
+                shutil.copystat(os.path.dirname(abs_src), parent_dir)
             shutil.copy2(abs_src, abs_dst, follow_symlinks=False)
             append_to_log(abs_dst)
 
@@ -195,23 +221,19 @@ def do_install(datafilename):
         run_install_script(d)
 
 def install_subdirs(d):
-    for (src_dir, inst_dir, dst_dir, mode, exclude) in d.install_subdirs:
-        if src_dir.endswith('/') or src_dir.endswith('\\'):
-            src_dir = src_dir[:-1]
-        src_prefix = os.path.join(src_dir, inst_dir)
-        print('Installing subdir %s to %s' % (src_prefix, dst_dir))
-        dst_dir = get_destdir_path(d, dst_dir)
-        d.dirmaker.makedirs(dst_dir, exist_ok=True)
-        do_copydir(d, src_prefix, src_dir, dst_dir, exclude)
-        dst_prefix = os.path.join(dst_dir, inst_dir)
-        set_mode(dst_prefix, mode)
+    for (src_dir, dst_dir, mode, exclude) in d.install_subdirs:
+        full_dst_dir = get_destdir_path(d, dst_dir)
+        print('Installing subdir %s to %s' % (src_dir, full_dst_dir))
+        d.dirmaker.makedirs(full_dst_dir, exist_ok=True)
+        do_copydir(d, src_dir, full_dst_dir, exclude)
+        set_mode(full_dst_dir, mode)
 
 def install_data(d):
     for i in d.data:
         fullfilename = i[0]
         outfilename = get_destdir_path(d, i[1])
         mode = i[2]
-        outdir = os.path.split(outfilename)[0]
+        outdir = os.path.dirname(outfilename)
         d.dirmaker.makedirs(outdir, exist_ok=True)
         print('Installing %s to %s' % (fullfilename, outdir))
         do_copyfile(fullfilename, outfilename)
@@ -221,7 +243,7 @@ def install_man(d):
     for m in d.man:
         full_source_filename = m[0]
         outfilename = get_destdir_path(d, m[1])
-        outdir = os.path.split(outfilename)[0]
+        outdir = os.path.dirname(outfilename)
         d.dirmaker.makedirs(outdir, exist_ok=True)
         print('Installing %s to %s' % (full_source_filename, outdir))
         if outfilename.endswith('.gz') and not full_source_filename.endswith('.gz'):
@@ -238,7 +260,7 @@ def install_man(d):
 def install_headers(d):
     for t in d.headers:
         fullfilename = t[0]
-        fname = os.path.split(fullfilename)[1]
+        fname = os.path.basename(fullfilename)
         outdir = get_destdir_path(d, t[1])
         outfilename = os.path.join(outdir, fname)
         print('Installing %s to %s' % (fname, outdir))
@@ -265,16 +287,10 @@ def run_install_script(d):
             rc = subprocess.call(script + args, env=child_env)
             if rc != 0:
                 sys.exit(rc)
-        except:
+        except OSError:
             print('Failed to run install script {!r}'.format(name))
             sys.exit(1)
 
-def is_elf_platform():
-    platname = platform.system().lower()
-    if platname == 'darwin' or platname == 'windows' or platname == 'cygwin':
-        return False
-    return True
-
 def check_for_stampfile(fname):
     '''Some languages e.g. Rust have output files
     whose names are not known at configure time.
@@ -304,7 +320,7 @@ def install_targets(d):
     for t in d.targets:
         fname = check_for_stampfile(t[0])
         outdir = get_destdir_path(d, t[1])
-        outname = os.path.join(outdir, os.path.split(fname)[-1])
+        outname = os.path.join(outdir, os.path.basename(fname))
         aliases = t[2]
         should_strip = t[3]
         install_rpath = t[4]
@@ -316,7 +332,7 @@ def install_targets(d):
             do_copyfile(fname, outname)
             if should_strip and d.strip_bin is not None:
                 if fname.endswith('.jar'):
-                    print('Not stripping jar target:', os.path.split(fname)[1])
+                    print('Not stripping jar target:', os.path.basename(fname))
                     continue
                 print('Stripping target {!r}'.format(fname))
                 ps, stdo, stde = Popen_safe(d.strip_bin + [outname])
@@ -332,7 +348,7 @@ def install_targets(d):
                 do_copyfile(pdb_filename, pdb_outname)
         elif os.path.isdir(fname):
             fname = os.path.join(d.build_dir, fname.rstrip('/'))
-            do_copydir(d, fname, os.path.dirname(fname), outdir, None)
+            do_copydir(d, fname, os.path.join(outdir, os.path.basename(fname)), None)
         else:
             raise RuntimeError('Unknown file type for {!r}'.format(fname))
         printed_symlink_error = False
@@ -350,10 +366,9 @@ def install_targets(d):
                     print("Symlink creation does not work on this platform. "
                           "Skipping all symlinking.")
                     printed_symlink_error = True
-        if is_elf_platform() and os.path.isfile(outname):
+        if os.path.isfile(outname):
             try:
-                e = depfixer.Elf(outname, False)
-                e.fix_rpath(install_rpath)
+                depfixer.fix_rpath(outname, install_rpath, False)
             except SystemExit as e:
                 if isinstance(e.code, int) and e.code == 0:
                     pass
@@ -366,7 +381,7 @@ def run(args):
         print('Installer script for Meson. Do not run on your own, mmm\'kay?')
         print('meson_install.py [install info file]')
     datafilename = args[0]
-    private_dir = os.path.split(datafilename)[0]
+    private_dir = os.path.dirname(datafilename)
     log_dir = os.path.join(private_dir, '../meson-logs')
     with open(os.path.join(log_dir, 'install-log.txt'), 'w') as lf:
         install_log_file = lf
